# -*- coding: utf-8 -*-

import os
import glob
from pprint import pprint
from paver.easy import *

__all__ = [
    'once',
    'load_paster_env', 'load_paster_app',
    'clyn', 'egg_info', 'package_map', 'develop',
    'upgrade_egg_cache', 'upload_dist',
]



NOT_RESULT = object()

def _once(func, *args, **kw):
    if func.__result is NOT_RESULT:
        func.__result = result = func(*args, **kw)
        return result
    else:
        return func.__result

def once(f):
    from decorator import decorator
    f.__result = NOT_RESULT
    return decorator(_once, f)



@task
def load_paster_env(options):
    #config_file, environ=os.environ):
    '''Virtual task for setting env from paster ini file'''
    import ConfigParser

    ini = os.path.abspath('%s.ini' % options.env)

    config = ConfigParser.RawConfigParser()
    config.optionxform = lambda s:s # we don't want lowercase env variables
    config.read(ini)

    if config.has_section('environ'):
        for option in config.options('environ'):
            value = config.get('environ', option)
            environ[option] = value



@task
def load_paster_app(options, ini=None):
    '''Virtual task for loading paster ini file'''
    from paste.deploy import loadapp
    if ini is None:
        env = options.get('env', os.environ.get('VINNETOU', 'development'))
        ini = os.path.abspath('%s.ini' % env)
    app = loadapp('config:%s' % ini)
    return app






def easy_install(pkg):
    from setuptools.command.easy_install import main
    main(['-UZ', pkg])



class package_metadata:

    def __init__(self, dir, metadata=None):
        self.dir = path(dir)
        if metadata is None:
            from cStringIO import StringIO
            from setuptools.sandbox import run_setup
            try:
                import json
            except ImportError:
                import simplejson as json
            olddir = os.getcwd()
            os.chdir(self.dir)
            try:
                out = sys.stdout
                sys.stdout = collector = StringIO()
                try:
                    run_setup('setup.py', ['-q', 'metadata_json'])
                finally:
                    sys.stdout = out
                data = collector.getvalue()
                metadata = json.loads(data)
            finally:
                os.chdir(olddir)
        self.__dict__.update(metadata)

    def develop(self, index_url=None, quiet=True):
        import subprocess
        cmd = [sys.executable, 'setup.py']
        if quiet:
            cmd.append('-q')
        cmd.append('develop')
        if index_url:
            cmd.append('-i')
            cmd.append(index_url)
        proc = subprocess.Popen(
            cmd,
            cwd=self.dir,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = proc.communicate()
        if proc.returncode != 0:
            print out
            print err
            sys.exit(1)

    def __repr__(self):
        return 'package(%s,%s)' % (self.egg_name, str(self.dir))


def find_package_dirs(directories):
    for dir in set([path(d).abspath() for d in directories]):
        if not dir.isdir():
            error('directory %s does not exist', dir)
            continue
        simple_setup = dir.joinpath('setup.py')
        if simple_setup.isfile():
            yield dir
        else:
            for s in dir.glob('*/setup.py'):
                yield s.dirname()


class package_working_set:

    def __init__(self, directories):
        self.directories = directories
        self.packages = {}
        for dir in directories:
            pm = package_metadata(dir)
            self.packages[pm.egg_name] = pm

    def __getitem__(self, name):
        return self.packages[name]

    def sorted_packages(self, all=False):
        pmap = {}
        for name, pkg in self.packages.items():
            pmap[name] = set(pkg.parsed_requirements.get('', []))
        extra_in_deps = reduce(set.union, pmap.values()) - set(pmap.keys())
        pmap.update(dict([(item, set()) for item in extra_in_deps]))
        while True:
            ordered = set(item for item, dep in pmap.items() if not dep)
            if not ordered:
                break
            for m in sorted(ordered):
                if all:
                    yield m
                elif m in self.packages:
                    yield m
            pmap = dict([(item, (dep - ordered))
                         for item, dep in pmap.items()
                         if item not in ordered])
        assert not pmap, "A cyclic dependency exists amongst %r" % pmap



def check_egg_info(dir):
    from setuptools.sandbox import run_setup
    cdir = os.getcwd()
    os.chdir(dir)
    try:
        run_setup('setup.py', ['-q', 'egg_info'])
    finally:
        if os.path.isdir('temp'):
            os.rmdir('temp')
        os.chdir(cdir)

@task
@consume_args
def egg_info(options):
    '''Run "python setup.py egg_info" for all packages'''
    for dir in sorted(find_package_dirs(options.args + ['.'])):
        info('egg_info(%s)', dir)
        check_egg_info(dir)


@task
@consume_args
def package_map(options):
    '''Prints package dependencies'''
    ws = package_working_set(find_package_dirs(options.args + ['.']))
    for name, pkg in sorted(ws.packages.items()):
        print name
        for req in sorted(pkg.parsed_requirements.get('', ())):
            print ' -', req


@task
@consume_args
def develop(options):
    '''Run "python setup.py develop" for all packages'''
    ws = package_working_set(find_package_dirs(options.args + ['.']))
    for name in ws.sorted_packages():
        info('develop(%s)', name)
        ws[name].develop(index_url=options.index_url)



@task
def upload_dist(options):
    '''Upload source distribution to egg cache host'''
    pkgdir = path(os.getcwd())
    for f in (pkgdir / 'dist').listdir():
        f.unlink()
    #sh('%s setup.py egg_info' % sys.executable, cwd=pkgdir)
    sh('%s setup.py sdist --format=zip' % sys.executable, cwd=pkgdir)
    eidir, = glob.glob(pkgdir / '*.egg-info')
    eidir = path(eidir)
    pkg_name = None
    for l in open(eidir / 'PKG-INFO'):
        n, v = l.split(None, 1)
        if n == 'Name:':
            pkg_name = v.strip()
            break
    if pkg_name is None:
        error('cannot resolve package name')
        sys.exit(1)

    for pyver in options.egg_cache.py_executables:
        sh('%s setup.py bdist_egg >/dev/null' % pyver, cwd=pkgdir)

    dest_dir = '%s/%s' % (options.egg_cache.dir, pkg_name)
    for f in (pkgdir / 'dist').listdir():
        sh('scp -q %s %s:%s' % (f, options.egg_cache.host, dest_dir))



@task
def clyn():
    '''Delete all *.pyc/*.pyo files from sources'''
    import subprocess
    cmd = '''find . -not ( -type d (
                -name .svn -o -name _svn -o -name .git -o
                -name .hg -o -name venv -o -name bin -o -name lib
             ) -prune ) -type f (
                -name *.pyc -o -name *.pyo
             ) -print'''.split()

    #print ' '.join(cmd)
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    for l in proc.stdout.readlines():
        l = l.strip()
        info('rm(%r)' % l)
        os.unlink(l)


@task
@consume_args
def upgrade_egg_cache(options):
    '''Upgrade given eggs on local egg cache server'''

    # XXX Problematicke pakaze:
    # - PyProtocols
    # - PEAK
    # - RuleDispatch
    # - Cython

    from setuptools import package_index
    from pkg_resources import Requirement

    cwd = path('.').abspath()
    tmpdir = cwd / 'tmp'

    #index = package_index.PackageIndex(index_url='http://ftp.uh.cz/eggs/')
    index = package_index.PackageIndex()
    info('scan(%s)' % index.index_url)
    index.scan_all()

    for pkg in options.args:

        tmpdir.rmtree()
        tmpdir.makedirs()

        req = Requirement.parse(pkg)
        dist = index.fetch_distribution(
                req, str(tmpdir), force_scan=False, source=True)
        if dist:
            info("")
            info("-----------------------------------------------------------")
            info(str(dist))
            info("-----------------------------------------------------------")
            zip = dist.location


            if zip.endswith('.tar.gz') or zip.endswith('.tgz'):
                sh('tar xzf %s' % zip, cwd=tmpdir)
            elif zip.endswith('.tar.bz2'):
                sh('tar xjf %s' % zip, cwd=tmpdir)
            elif zip.endswith('.zip'):
                sh('unzip -q %s' % zip, cwd=tmpdir)
            else:
                error('unknown package format %s' % zip)
                sys.exit(1)

            def run_setup(exe, cwd, *args):
                args = ", ".join(["'%s'" % a for a in args])
                try:
                    sh(('%s -c "from setuptools.sandbox import run_setup; '
                         ' run_setup(\'setup.py\', [%s])" >/dev/null') % (
                             exe, args),
                        cwd=cwd)
                except:
                    error('build for %s failed', cwd)

            pkg_files = []
            for pkgroot in tmpdir.listdir():
                if not pkgroot.isdir():
                    continue
                info(str(pkgroot))
                run_setup(sys.executable, pkgroot,
                          'sdist', '--quiet', '--format=zip')
                for venv in sorted(glob.glob(cwd / 'venv*')):
                    venv = path(venv)
                    run_setup(venv/'bin'/'python', pkgroot,
                              'bdist_egg', '--quiet')
                pkg_files.extend((pkgroot / 'dist').listdir())
                break

            dest_dir = '%s/%s' % (options.egg_cache.dir, dist.project_name)
            sh('ssh %s "mkdir -p %s"' % (options.egg_cache.host, dest_dir))
            for f in pkg_files:
                sh('scp -q %s %s:%s' % (
                    f, options.egg_cache.host, dest_dir))

        else:
            error('package %s not found' % pkg)

