#!/usr/bin/env python

import os
import logging
import time
import sys

from vinnetou.base import make_wsgi_app
from vinnetou.utils import sane_environ
from vinnetou.middleware import *

DEFAULT_PORT = 4000


def get_opts():
    from optparse import OptionParser
    usage = 'usage: %prog [options] appname'
    parser = OptionParser(usage=usage)
    parser.add_option(
        '--start', action="store_const", dest="action",
        help="start application", const="start"
    )
    parser.add_option(
        '--stop', action="store_const", dest="action",
        help="stop application", const="stop"
    )
    parser.add_option(
        '--restart', action="store_const", dest="action",
        help="restart application", const="restart"
    )
    parser.add_option(
        '-p', '--port', type="int",
        help="port [default: %default]", default=DEFAULT_PORT,
    )
    parser.add_option(
        '-l', '--listen', metavar='ADDRESS',
        help="listen on address [default: %default]", default='localhost',
    )
    parser.add_option(
        '-d', '--debug', action="store_true",
        help="do not run as daemon", default=False,
    )
    parser.add_option(
        '-c', '--children', type="int",
        help="maximal number or children [default: %default]", default=4,
    )
    parser.add_option(
        '--minspare', type="int",
        help="minimal number or spare children [default: %default]", default=1,
    )
    parser.add_option(
        '--maxspare', type="int",
        help="maximal number or spare children [default: %default]", default=4,
    )
    parser.add_option(
        '-r', '--requests', type="int", default=50,
        help="maximum number of requests per children [default: %default]",
    )
    parser.add_option(
        '-i', '--ini', metavar="INIFILE", action="append",
        help="additional ini files", default=[]
    )
    parser.add_option(
        '--log-level', metavar="LEVELNAME", default=None,
        help="log level [default: ERROR or DEBUG with -d]"
    )
    parser.add_option(
        '--log-file', default=None, help="log file"
    )
    parser.add_option(
        '--log-smtp', metavar="SENDER,RECIPIENT,SUBJECT", default=None,
        help="log via SMTP"
    )
    parser.add_option(
        '--log-smtp-host', metavar="HOST[:PORT]", default='localhost',
        help="SMTP host [default: %default]",
    )
    parser.add_option(
        '-k', '--lock-file',  help="lock file [default=%default]",
        default='/var/lock/scgid.%(port)d.lck'
    )
    parser.add_option(
        '--kill-delay', type="int", default=20,
        help="delay before kill -9 when stopping [default: %default]",
    )
    parser.add_option(
        '--rlimit-as', type="int", default=-1, metavar="BYTES",
        help="maximum size of process's virtual memory (address space)",
    )
    parser.add_option(
        '--rlimit-data', type="int", default=-1, metavar="BYTES", help=(
            "maximum  size  of  the  process's data segment "
            "(initialized data, uninitialized data, and heap)"),
    )
    parser.add_option(
        '--rlimit-rss', type="int", default=-1, metavar="PAGES", help= (
            "limit  (in  pages) of the process's resident set "
            "(the number of virtual pages resident in RAM)"),
    )
    parser.add_option(
        '--rlimit-stack', type="int", default=-1, metavar="BYTES",
        help="maximum size of process stack in bytes",
    )
    opts, args = parser.parse_args()
    if len(args) < 1:
        parser.error('application name required')
    else:
        opts.app_name = args[0]
    if not opts.action:
        opts.action = "start"
    opts.lock_file = opts.lock_file % {'port': opts.port}
    if not opts.log_level:
        opts.log_level = opts.debug and 'DEBUG' or 'ERROR'
    opts._log_level = getattr(logging, opts.log_level)
    set_rlimits(opts)
    return opts


def set_rlimits(opts):
    import resource
    for lim in 'as', 'data', 'rss', 'stack':
        val = getattr(opts, 'rlimit_%s' % lim)
        if val != -1:
            resource.setrlimit(
                getattr(resource, 'RLIMIT_%s' % lim.upper()),
                (val, val)
            )


def setup_logging(opts):
    from flup.server import scgi_base as base
    logger = logging.getLogger(base.LoggerName)
    if not opts.debug:
        for h in list(logger.handlers):
            if isinstance(h, logging.StreamHandler):
                logger.removeHandler(h)
    if opts.log_file:
        fileHandler = logging.FileHandler(opts.log_file)
        fileHandler.setFormatter(
            logging.Formatter('%(asctime)s : %(message)s', '%Y-%m-%d %H:%M:%S'))
        logger.addHandler(fileHandler)
    if opts.log_smtp:
        from logging.handlers import SMTPHandler
        sender,recipient,subject = [p.strip() for p in opts.log_smtp.split(',')]
        smtpHandler = SMTPHandler(opts.log_smtp_host,sender,recipient,subject)
        logger.addHandler(smtpHandler)
    return logger


def do_start(opts):
    from lib4.procutils import ProcessInstanceLock, daemonize
    from flup.server.scgi_fork import WSGIServer

    if not opts.debug:
        daemonize()
    try:
        from lib4.remote_console import setup_remote_console
        setup_remote_console()
    except:
        pass
    logger = setup_logging(opts)
    lck = ProcessInstanceLock(opts.lock_file)
    if not lck.acquire(blocking=False):
        logger.error('another instance is running')
        sys.exit(1)
    env = os.environ.copy()
    env['APP_NAME'] = opts.app_name
    app = make_wsgi_app(env)
    app = proctitle_middleware_factory(app, env)
    srv = WSGIServer(
        app,
        maxChildren   = opts.children,
        maxRequests   = opts.requests,
        minSpare      = opts.minspare,
        maxSpare      = opts.maxspare,
        bindAddress   = (opts.listen, opts.port),
        loggingLevel  = opts._log_level,
        debug         = opts.debug,
    )
    srv.run()


def do_stop(opts):
    from signal import SIGTERM, SIGKILL
    try:
        pid = int(file(opts.lock_file).read().strip())
    except IOError:
        print "Could not stop, pid file %r missing." % opts.lock_file
        return 1
    sys.stdout.write('Stopping %s[%d]: ' % (opts.app_name, pid))
    sys.stdout.flush()
    try:
        os.kill(pid, SIGTERM)
    except OSError, msg:
        print 'Cannot stop process %d: %s' % (pid, msg)
    else:
        exited = False
        for i in range(opts.kill_delay*3):
            time.sleep(0.3)
            try:
                os.kill(pid, 0)
            except:
                exited = True
                break
            sys.stdout.write('.')
            sys.stdout.flush()
        if exited:
            sys.stdout.write("OK\n")
        else:
            sys.stdout.write("Killed\n")
            try:
                os.kill(pid, SIGKILL)
            except:
                pass
        try:
            os.remove(opts.lock_file)
        except:
            pass
    return 0


def do_restart(opts):
    do_stop(opts)
    do_start(opts)


def main():
    opts = get_opts()
    meth = globals()['do_' + opts.action]
    meth(opts)

if __name__ == '__main__':
    main()
